#ifndef SERIAL_H
#define SERIAL_H

#include <stdint.h>
#include <string.h>

#define host8le           (uint8_t)
#ifdef BIGENDIAN
  #define host16le        byteswap16
  #define host32le        byteswap32
  #define host64le        byteswap64
#else
  #define host16le        (uint16_t)
  #define host32le        (uint32_t)
  #define host64le        (uint64_t)
#endif

#define __DEF_READ(NAME_PFIX,TYPE,WIDTH)\
  inline void read_##NAME_PFIX##WIDTH(uint8_t const *buf,TYPE *data);\
  inline void read_##NAME_PFIX##WIDTH(uint8_t const *buf,TYPE *data){\
    TYPE tmp;\
    memcpy(&tmp,buf,sizeof(TYPE));\
    *data = host##WIDTH##le(tmp);\
  }

#define __DEF_WRITE(NAME_PFIX,TYPE,WIDTH)\
  inline void write_##NAME_PFIX##WIDTH(uint8_t *buf,TYPE const *data);\
  inline void write_##NAME_PFIX##WIDTH(uint8_t *buf,TYPE const *data){\
    TYPE tmp;\
    tmp = host##WIDTH##le(*data);\
    memcpy(buf,&tmp,sizeof(TYPE));\
  }

#define _DEF_READWRITE(WIDTH)\
  __DEF_READ  (u,uint##WIDTH##_t,WIDTH)\
  __DEF_WRITE (u,uint##WIDTH##_t,WIDTH)\

_DEF_READWRITE(8)
_DEF_READWRITE(16)
_DEF_READWRITE(32)
_DEF_READWRITE(64)

#undef __DEF_READ
#undef __DEF_WRITE
#undef _DEF_READWRITE

inline uint16_t     byteswap16(uint16_t n);
inline uint32_t     byteswap32(uint32_t n);
inline uint64_t     byteswap64(uint64_t n);

inline uint16_t byteswap16(uint16_t n){
#ifdef __GNUC__
  return __builtin_bswap16(n);
#else
  return n>>8 | n<<8;
#endif
}

inline uint32_t byteswap32(uint32_t n){
#ifdef __GNUC__
  return __builtin_bswap32(n);
#else
  return (n              ) >> 24 |
         (n & 0x00FF0000u) >>  8 | 
         (n & 0x0000FF00u) <<  8 | 
         (n              ) << 24;
#endif
}

inline uint64_t byteswap64(uint64_t n){
#ifdef __GNUC__
  return __builtin_bswap64(n);
#else
  return (n                        ) >> 56 |
         (n & 0x00FF000000000000ull) >> 40 | 
         (n & 0x0000FF0000000000ull) >> 24 | 
         (n & 0x000000FF00000000ull) >>  8 | 
         (n & 0x00000000FF000000ull) <<  8 | 
         (n & 0x0000000000FF0000ull) << 24 | 
         (n & 0x000000000000FF00ull) << 40 | 
         (n                        ) << 56;
#endif
}

#endif

